from typing import Any

from qxmt.configs import DeviceConfig
from qxmt.devices.amazon import AMAZON_BRAKET_DEVICES
from qxmt.devices.base import BaseDevice
from qxmt.devices.ibmq import IBMQ_REAL_DEVICES
from qxmt.exceptions import InvalidPlatformError
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


class DeviceBuilder:
    """
    Builder class for quantum devices. This class abstracts the instantiation of quantum device objects
    across multiple platforms (e.g., PennyLane, IBMQ, Amazon Braket) based on a unified configuration.

    This builder enables users to create a device instance by simply providing a DeviceConfig, without
    worrying about the underlying platform-specific details.

    Args:
        config (DeviceConfig): Configuration for the quantum device. This includes platform, device name,
            backend name, number of qubits, shots, and random seed.
        logger (Any): Logger instance for logging.

    Methods:
        build() -> BaseDevice:
            Instantiates and returns a quantum device object corresponding to the specified platform and
            device name.

    Returns:
        BaseDevice: An instance of a quantum device suitable for the specified platform and configuration.

    Example:
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     device_name="default.qubit",
        ...     backend_name=None,
        ...     n_qubits=2,
        ...     shots=1000,
        ...     random_seed=42,
        ... )
        >>> device = DeviceBuilder(config).build()
    """

    def __init__(
        self,
        config: DeviceConfig,
        logger: Any = LOGGER,
    ) -> None:
        """Initialize the device builder.

        Args:
            config (DeviceConfig): Configuration for the quantum device. It is element of the ExperimentConfig.
        """
        self.config: DeviceConfig = config
        self.logger: Any = logger

    def build(self) -> BaseDevice:
        """Build a quantum device. it can be a general-purpose device overseeing multiple platforms.

        Returns:
            BaseDevice: General-purpose device overseeing multiple platforms
        """
        platform = self.config.platform
        device_name = self.config.device_name
        backend_name = self.config.backend_name
        n_qubits = self.config.n_qubits
        shots = self.config.shots
        random_seed = self.config.random_seed

        if platform == "pennylane":
            if device_name in IBMQ_REAL_DEVICES:
                from qxmt.devices.ibmq_device import IBMQDevice

                return IBMQDevice(platform, device_name, backend_name, n_qubits, shots, random_seed, self.logger)
            elif device_name in AMAZON_BRAKET_DEVICES:
                from qxmt.devices.amazon_device import AmazonBraketDevice

                return AmazonBraketDevice(
                    platform, device_name, backend_name, n_qubits, shots, random_seed, self.logger
                )
            else:
                from qxmt.devices.pennylane_device import PennyLaneDevice

                return PennyLaneDevice(platform, device_name, backend_name, n_qubits, shots, random_seed, self.logger)
        else:
            raise InvalidPlatformError(f'"{platform}" is not implemented.')



from typing import cast

from pennylane.workflow.qnode import SupportedDiffMethods

from qxmt.ansatze import AnsatzBuilder, BaseAnsatz
from qxmt.configs import ExperimentConfig
from qxmt.constants import DEFAULT_N_JOBS
from qxmt.devices import BaseDevice, DeviceBuilder
from qxmt.exceptions import InvalidModelNameError
from qxmt.hamiltonians import BaseHamiltonian, HamiltonianBuilder
from qxmt.models.vqe.base import BaseVQE
from qxmt.models.vqe.basic import BasicVQE


class VQEModelBuilder:
    """Builder class for VQE models."""

    def __init__(self, config: ExperimentConfig, n_jobs: int = DEFAULT_N_JOBS) -> None:
        """Initialize the model builder."""
        self.config: ExperimentConfig = config
        self.n_jobs: int = n_jobs
        self.device: BaseDevice
        self.hamiltonian: BaseHamiltonian
        self.ansatz: BaseAnsatz
        self.model: BaseVQE

    def _set_device(self) -> None:
        device_config = self.config.device
        self.device = DeviceBuilder(config=device_config).build()

    def _set_hamiltonian(self) -> None:
        hamiltonian_config = self.config.hamiltonian
        if hamiltonian_config is None:
            raise ValueError("Hamiltonian config is not provided.")

        self.hamiltonian = HamiltonianBuilder(config=hamiltonian_config).build()

    def _set_ansatz(self) -> None:
        ansatz_config = self.config.ansatz
        if ansatz_config is None:
            raise ValueError("Ansatz config is not provided.")

        self.ansatz = AnsatzBuilder(config=ansatz_config, device=self.device, hamiltonian=self.hamiltonian).build()

    def _set_model(self) -> None:
        match self.config.model.name:
            case "basic":
                self.model = BasicVQE(
                    device=self.device,
                    hamiltonian=self.hamiltonian,
                    ansatz=self.ansatz,
                    diff_method=cast(SupportedDiffMethods, self.config.model.diff_method),
                    max_steps=self.config.model.params.get("max_steps", 20),
                    min_steps=self.config.model.params.get("min_steps", None),
                    tol=self.config.model.params.get("tol", 1e-6),
                    verbose=self.config.model.params.get("verbose", True),
                    optimizer_settings=self.config.model.optimizer_settings,
                    init_params_config=self.config.model.params.get("init_params", None),
                )
            case _:
                raise InvalidModelNameError(f'"{self.config.model.name}" is not implemented.')

    def build(self) -> BaseVQE:
        """Build the VQE model."""
        self._set_device()
        self._set_hamiltonian()
        self._set_ansatz()
        self._set_model()

        return self.model



from qxmt.evaluation.metrics.base import BaseMetric

__all__ = ["BaseMetric"]

from qxmt.evaluation.metrics.defaults_classification import (
    Accuracy,
    F1Score,
    Precision,
    Recall,
)
from qxmt.evaluation.metrics.defaults_regression import (
    MeanAbsoluteError,
    R2Score,
    RootMeanSquaredError,
)
from qxmt.evaluation.metrics.defaults_vqe import FCIEnergy, FinalCost, HFEnergy

__all__ += [
    "Accuracy",
    "Recall",
    "Precision",
    "F1Score",
    "MeanAbsoluteError",
    "RootMeanSquaredError",
    "R2Score",
    "FCIEnergy",
    "FinalCost",
    "HFEnergy",
]



from typing import Callable, Literal, cast

import numpy as np
import pennylane as qml
from pennylane.measurements.sample import SampleMP
from pennylane.measurements.state import StateMP
from rich.progress import track

from qxmt.devices import BaseDevice
from qxmt.exceptions import ModelSettingError
from qxmt.feature_maps import BaseFeatureMap
from qxmt.kernels import BaseKernel
from qxmt.kernels.sampling import sample_results_to_probs


class ProjectedKernel(BaseKernel):
    """Projected kernel class.
    The projected kernel is a quantum kernel that projects the quantum state to a specific basis
    and computes the kernel value based on the projected measurement results.
    Reference: https://www.nature.com/articles/s41467-021-22539-9

    Args:
        BaseKernel (_type_): base class of kernel

    Examples:
        >>> import numpy as np
        >>> from qxmt.kernels.pennylane.projected_kernel import ProjectedKernel
        >>> from qxmt.feature_maps.pennylane.defaults import ZZFeatureMap
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1024,
        >>> )
        >>> device = DeviceBuilder(config).build()
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> kernel = ProjectedKernel(device, feature_map)
        >>> x1 = np.random.rand(2)
        >>> x2 = np.random.rand(2)
        >>> kernel.compute(x1, x2)
        0.86
    """

    def __init__(
        self,
        device: BaseDevice,
        feature_map: BaseFeatureMap | Callable[[np.ndarray], None],
        gamma: float = 1.0,
        projection: Literal["x", "y", "z"] = "z",
    ) -> None:
        """Initialize the ProjectedKernel class.

        Args:
            device (BaseDevice): device instance for quantum computation
            feature_map (BaseFeatureMap | Callable[[np.ndarray], None]): feature map instance or function
            gamma (float): gamma parameter for kernel computation
            projection (str): projection method for kernel computation
        """
        if projection not in ["x", "y", "z"]:
            raise ValueError("Projection method must be 'x', 'y', or 'z'.")

        super().__init__(device, feature_map)
        self.n_qubits = device.n_qubits
        self.gamma = gamma
        self.projection = projection
        self.qnode = None
        self.state_memory = {}

    def _initialize_qnode(self) -> None:
        if self.qnode is None:
            self.qnode = qml.QNode(self._circuit, device=self.device.get_device(), cache=False)

    def _process_measurement_results(self, results: list[float]) -> np.ndarray:
        """Process the measurement results based on the projection method.

        Args:
            results (list[float]): projected measurement results for each qubit

        Returns:
            np.ndarray: processed measurement results
        """
        if self.projection == "xyz_sum":
            sum_results: list[float] = []
            idx_size = 3  # x, y, z
            for i in range(0, self.n_qubits, idx_size):
                sum_result: float = np.sum([result for result in results[i : i + idx_size]])
                sum_results.append(sum_result)
            projected_results = np.array(sum_results)
        else:
            projected_results = np.array(results)

        return projected_results

    def _calculate_expected_values_by_z(self, probs: np.ndarray, target_qubit: int) -> float:
        mask = 1 << target_qubit
        expval_z = 0.0
        for i, prob in enumerate(probs):
            if (i & mask) == 0:
                expval_z += prob
            else:
                expval_z -= prob
        return expval_z

    def _calculate_expected_values(self, probs: np.ndarray) -> np.ndarray:
        # calculate the expected values by Z basis for each qubit
        # when the projection method is "x" or "y", apply Hadamard or RY gate in _circuit method
        projected_exp_value = np.array([self._calculate_expected_values_by_z(probs, i) for i in range(self.n_qubits)])

        return projected_exp_value

    def _circuit(self, x: np.ndarray) -> SampleMP | list[SampleMP] | StateMP:
        if self.feature_map is None:
            raise ModelSettingError("Feature map must be provided for FidelityKernel.")

        self.feature_map(x)

        # apply projection operators for calculating the expected values by Z basis
        if self.projection == "x":
            for i in range(self.n_qubits):
                qml.Hadamard(wires=i)
        elif self.projection == "y":
            for i in range(self.n_qubits):
                qml.RY(qml.numpy.array(np.pi / 2), wires=i)

        if (self.is_sampling) and (self.device.is_amazon_device()):
            # Amazon Braket does not support directry sample by computational basis
            return [qml.sample(op=qml.PauliZ(wires=i)) for i in range(self.n_qubits)]
        elif self.is_sampling:
            return qml.sample(wires=self.n_qubits)
        else:
            return qml.state()

    def _compute_matrix_by_state_vector(
        self, x1_array: np.ndarray, x2_array: np.ndarray, bar_label: str = "", show_progress: bool = True
    ) -> np.ndarray:
        """Compute the kernel matrix based on the state vector.
        This method is only available in the non-sampling mode.
        Each kernel value computed by theoritically probability distribution by state vector.

        Args:
            x1_array (np.ndarray): numpy array representing the all data points (ex: Train data)
            x2_array (np.ndarray): numpy array representing the all data points (ex: Train data, Test data)
            bar_label (str): label for progress bar
            show_progress (bool): flag for showing progress bar

        Returns:
            np.ndarray: computed kernel matrix
        """
        self._initialize_qnode()
        if self.qnode is None:
            raise RuntimeError("QNode is not initialized.")

        unique_inputs = set([tuple(x) for x in x1_array] + [tuple(x) for x in x2_array])
        if show_progress:
            bar_label = f" ({bar_label})" if bar_label else ""
            iterator = track(unique_inputs, description=f"Computing Kernel Matrix{bar_label}")
        else:
            iterator = unique_inputs

        for x_tuple in iterator:
            if x_tuple not in self.state_memory:
                x_state = self.qnode(np.array(x_tuple))
                self.state_memory[x_tuple] = self._calculate_expected_values(np.abs(x_state) ** 2)

        kernel_matrix = np.zeros((len(x1_array), len(x2_array)))
        for i, x1 in enumerate(x1_array):
            for j, x2 in enumerate(x2_array):
                kernel_matrix[i, j] = np.exp(
                    -self.gamma * np.sum((self.state_memory[tuple(x1)] - self.state_memory[tuple(x2)]) ** 2)
                )

        return kernel_matrix

    def _compute_by_sampling(self, x1: np.ndarray, x2: np.ndarray) -> tuple[float, np.ndarray]:
        """Compute the projected kernel value between two data points.
        This method is only available in the sampling mode.
        Each kernel value computed by sampling the quantum circuit.

        Args:
            x1 (np.ndarray): numpy array representing the first data point
            x2 (np.ndarray): numpy array representing the second data point

        Returns:
            tuple[float, np.ndarray]: projected kernel value and probability distribution
        """
        self._initialize_qnode()
        if self.qnode is None:
            raise RuntimeError("QNode is not initialized.")

        x1_result = self.qnode(x1)
        x2_result = self.qnode(x2)

        if (self.is_sampling) and (self.device.is_amazon_device()):
            # PauliZ basis convert to computational basis (-1->1, 1->0)
            x1_binary_result = (np.array(x1_result).T == -1).astype(int)
            x2_binary_result = (np.array(x2_result).T == -1).astype(int)
            # convert the sample results to probability distribution
            # shots must be over 0 when sampling mode
            x1_probs = sample_results_to_probs(x1_binary_result, self.n_qubits, cast(int, self.device.shots))
            x2_probs = sample_results_to_probs(x2_binary_result, self.n_qubits, cast(int, self.device.shots))
        else:
            # convert the sample results to probability distribution
            # shots must be over 0 when sampling mode
            x1_probs = sample_results_to_probs(x1_result, self.n_qubits, cast(int, self.device.shots))
            x2_probs = sample_results_to_probs(x2_result, self.n_qubits, cast(int, self.device.shots))

        # compute expected values for projection operators
        x1_projected = self._calculate_expected_values(x1_probs)
        x2_projected = self._calculate_expected_values(x2_probs)

        # compute gaussian kernel value based on the projected measurement results
        kernel_value = np.exp(-self.gamma * np.sum((x1_projected - x2_projected) ** 2))

        return kernel_value, (x1_probs + x2_probs) / 2



from typing import Any, Optional, Type, get_args

import pandas as pd

from qxmt.evaluation.metrics.base import BaseMetric
from qxmt.evaluation.metrics.defaults_classification import (
    DEFAULT_CLF_METRICS_NAME,
    NAME2CLF_METRIC,
)
from qxmt.evaluation.metrics.defaults_regression import (
    DEFAULT_REG_METRICS_NAME,
    NAME2REG_METRIC,
)
from qxmt.evaluation.metrics.defaults_vqe import (
    DEFAULT_VQE_METRICS_NAME,
    NAME2VQE_METRIC,
)
from qxmt.utils import load_object_from_yaml


class Evaluation:
    """A class for evaluating model performance using various metrics.

    This class provides functionality to evaluate model predictions against actual values
    using both default and custom metrics. Results can be accessed as dictionaries or pandas DataFrames.

    Attributes:
        DEFAULT_METRICS_NAME (list[str]): List of default metric names
        NAME2METRIC (dict[str, Type[BaseMetric]]): Mapping of metric names to metric classes

    Examples:
        >>> from qxmt.evaluation.evaluation import Evaluation
        >>> params = {"actual": np.array([1, 0, 1]), "predicted": np.array([1, 1, 1])}
        >>> evaluation = Evaluation(params)
        >>> evaluation.evaluate()
        >>> evaluation.to_dict()
        {'accuracy': 0.6666666666666666, 'precision': 0.6666666666666666, 'recall': 1.0, 'f1_score': 0.8}
        >>> evaluation.to_dataframe()
           accuracy  precision  recall  f1_score
        0  0.666667   0.666667     1.0       0.8
    """

    DEFAULT_METRICS_NAME: list[str] = []
    NAME2METRIC: dict[str, Type[BaseMetric]] = {}

    def __init__(
        self,
        params: dict[str, Any],
        default_metrics_name: Optional[list[str]] = None,
        custom_metrics: Optional[list[dict[str, Any]]] = None,
    ) -> None:
        """Initialize the evaluation class with parameters and metrics.

        Args:
            params (dict[str, Any]): Dictionary containing evaluation parameters
            default_metrics_name (Optional[list[str]], optional):
                List of default metric names to use. If None, uses class default metrics.
            custom_metrics (Optional[list[dict[str, Any]]], optional):
                List of custom metric configurations. Each metric must be a subclass of BaseMetric.
        """
        self.params: dict[str, Any] = params
        self.default_metrics_name: list[str]
        self.custom_metrics_name: list[str]
        self.default_metrics: list[BaseMetric]
        self.custom_metrics: list[BaseMetric]

        self.init_default_metrics(default_metrics_name)
        self.init_custom_metrics(custom_metrics)

    def __repr__(self) -> str:
        return (
            f"Evaluation(params={self.params}, "
            f"default_metrics_name={self.default_metrics_name})"
            f"custom_metrics_name={self.custom_metrics_name}"
        )

    def init_default_metrics(self, default_metrics_name: Optional[list[str]]) -> None:
        """Initialize and validate default metrics.

        Args:
            default_metrics_name (Optional[list[str]]): List of default metric names to use

        Raises:
            ValueError: If any specified metric is not implemented
        """
        if default_metrics_name is not None:
            self.default_metrics_name = default_metrics_name
        else:
            self.default_metrics_name = self.DEFAULT_METRICS_NAME

        self.default_metrics = []
        for metric_name in self.default_metrics_name:
            if metric_name not in self.NAME2METRIC:
                raise ValueError(f"{metric_name} is not implemented.")

            metric = self.NAME2METRIC[metric_name](name=metric_name)
            self.default_metrics.append(metric)

    def init_custom_metrics(self, custom_metrics: Optional[list[dict[str, Any]]]) -> None:
        """Initialize and validate custom metrics.

        Args:
            custom_metrics (Optional[list[dict[str, Any]]]): List of custom metric configurations

        Raises:
            ValueError: If any custom metric is not a subclass of BaseMetric
        """
        self.custom_metrics_name = []
        self.custom_metrics = []

        if custom_metrics is not None:
            for metric_config in custom_metrics:
                metric = load_object_from_yaml(config=metric_config)
                if not isinstance(metric, BaseMetric):
                    raise ValueError("Custom metrics must be a subclass of BaseMetric.")

                self.custom_metrics.append(metric)
                self.custom_metrics_name.append(metric.name)

    def set_evaluation_result(self, metrics: list[BaseMetric]) -> None:
        """Calculate scores for a list of metrics using the evaluation parameters.

        This method evaluates each metric using only the parameters it requires.
        If a metric requires parameters that are not available in self.params,
        a KeyError will be raised.

        Args:
            metrics (list[BaseMetric]): List of metrics to evaluate

        Raises:
            KeyError: If a required parameter is missing from self.params
        """
        for metric in metrics:
            metric.set_score(**self.params)

    def evaluate(self) -> None:
        """Evaluate all default and custom metrics."""
        self.set_evaluation_result(self.default_metrics)
        self.set_evaluation_result(self.custom_metrics)

    def to_dict(self) -> dict:
        """Convert evaluation results to a dictionary.

        Returns:
            dict: Dictionary containing metric names as keys and their scores as values

        Raises:
            ValueError: If metrics have not been evaluated yet
        """
        metrics = self.default_metrics + self.custom_metrics
        for metric in metrics:
            if metric.score is None:
                raise ValueError("Metrics are not evaluated yet.")

        data = {metric.name: metric.score for metric in metrics}
        return data

    def to_dataframe(
        self,
        id: Optional[str] = None,
        id_columns_name: Optional[str] = None,
    ) -> pd.DataFrame:
        """Convert evaluation results to a pandas DataFrame.

        Args:
            id (Optional[str], optional): Identifier for the evaluation (e.g., run_id)
            id_columns_name (Optional[str], optional): Name of the ID column in the DataFrame

        Returns:
            pd.DataFrame: DataFrame containing evaluation metrics
        """
        data = self.to_dict()
        df = pd.DataFrame(data, index=[0])
        if (id is not None) and (id_columns_name is not None):
            df.insert(0, id_columns_name, id)

        return df


class ClassificationEvaluation(Evaluation):
    """Evaluation class specifically for classification tasks.

    Inherits from Evaluation and uses classification-specific metrics.
    """

    DEFAULT_METRICS_NAME = list(get_args(DEFAULT_CLF_METRICS_NAME))
    NAME2METRIC = NAME2CLF_METRIC


class RegressionEvaluation(Evaluation):
    """Evaluation class specifically for regression tasks.

    Inherits from Evaluation and uses regression-specific metrics.
    """

    DEFAULT_METRICS_NAME = list(get_args(DEFAULT_REG_METRICS_NAME))
    NAME2METRIC = NAME2REG_METRIC


class VQEEvaluation(Evaluation):
    """Evaluation class specifically for Variational Quantum Eigensolver (VQE) tasks.

    Inherits from Evaluation and uses VQE-specific metrics.
    """

    DEFAULT_METRICS_NAME = list(get_args(DEFAULT_VQE_METRICS_NAME))
    NAME2METRIC = NAME2VQE_METRIC



from enum import Enum, auto
from logging import Logger
from typing import Literal, Optional

import pennylane as qml
from pennylane import numpy as qnp
from pennylane.ops.op_math import Sum

from qxmt.constants import PENNYLANE_PLATFORM
from qxmt.hamiltonians import BaseHamiltonian
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)

DATA_MODULE_NAME = "qchem"
SUPPORTED_BASIS_NAMES = Literal["STO-3G", "6-31G", "6-311G", "CC-PVDZ"]
SUPPORTED_UNITS = Literal["angstrom", "bohr"]
SUPPORTED_METHODS = Literal["dhf", "pyscf", "openfermion"]
SUPPORTED_MAPPINGS = Literal["jordan_wigner", "bravyi_kitaev", "parity"]


class InitializationType(Enum):
    DATASET = auto()
    DIRECT_MOLECULE = auto()
    INVALID = auto()


class MolecularHamiltonian(BaseHamiltonian):
    """Molecular Hamiltonian for quantum chemistry calculations.

    This class represents a molecular Hamiltonian using PennyLane's quantum chemistry module.
    It supports both full and active space calculations for molecular systems.

    Args:
        molname: Name of the molecule in PennyLane's dataset.
        bondlength: Bond length of the molecule in Angstroms.
        symbols: List of atomic symbols (e.g., ['H', 'H'] for H2).
        coordinates: Array of atomic coordinates in Angstroms.
        charge: Total charge of the molecule. Defaults to 0.
        multi: Multiplicity of the molecule. Defaults to 1.
        basis_name: Basis set name. Only supported ["sto-3g", "6-31g", "6-311g", "cc-pvdz"]. Defaults to "sto-3g".
        unit: Unit of the coordinates. Only supported ["angstrom", "bohr"]. Defaults to "angstrom".
        method: Method to use for the calculation. Only supported ["dhf", "pyscf", "openfermion"]. Defaults to "dhf".
        active_electrons: Number of active electrons. If None, uses all electrons.
        active_orbitals: Number of active orbitals. If None, uses all orbitals.
        mapping: Mapping to use for the calculation. Defaults to "jordan_wigner".

    Attributes:
        hamiltonian: PennyLane Hamiltonian operator.
        n_qubits: Number of qubits required for the simulation.
        molecule: PennyLane Molecule object.
    """

    def __init__(
        self,
        molname: Optional[str] = None,
        bondlength: Optional[float | str] = None,
        symbols: Optional[list[str]] = None,
        coordinates: Optional[qnp.tensor | list[float] | list[list[float]]] = None,
        charge: int = 0,
        multi: int = 1,
        basis_name: SUPPORTED_BASIS_NAMES = "STO-3G",
        unit: SUPPORTED_UNITS = "angstrom",
        method: SUPPORTED_METHODS = "dhf",
        active_electrons: Optional[int] = None,
        active_orbitals: Optional[int] = None,
        mapping: SUPPORTED_MAPPINGS = "jordan_wigner",
        logger: Logger = LOGGER,
    ) -> None:
        super().__init__(platform=PENNYLANE_PLATFORM)

        self.molname: Optional[str] = molname
        self.bondlength: Optional[float | str] = bondlength
        self.symbols: Optional[list[str]] = symbols
        self.coordinates: Optional[qnp.tensor | list[float] | list[list[float]]] = (
            qnp.array(coordinates, requires_grad=False) if isinstance(coordinates, list) else coordinates
        )  # type: ignore
        self.charge: int = charge
        self.multi: int = multi
        self.basis_name: SUPPORTED_BASIS_NAMES = basis_name
        self.unit: SUPPORTED_UNITS = unit
        self.method: SUPPORTED_METHODS = method
        self.active_electrons: Optional[int] = active_electrons
        self.active_orbitals: Optional[int] = active_orbitals
        self.mapping: SUPPORTED_MAPPINGS = mapping
        self.logger: Logger = logger
        self.hamiltonian: Sum
        self.n_qubits: int
        self.molecule: qml.qchem.Molecule
        self.hf_energy: float
        self.fci_energy: Optional[float] = None
        self._dataset: list = []

        self._initialize_hamiltonian()
        self._set_hf_energy()
        self._set_fci_energy()

    def _determine_initialization_type(self) -> InitializationType:
        """Determine the initialization type based on the provided parameters.

        Returns:
            InitializationType: The type of initialization.
        """
        if self.molname is not None and self.bondlength is not None:
            return InitializationType.DATASET
        elif self.symbols is not None and self.coordinates is not None:
            return InitializationType.DIRECT_MOLECULE
        return InitializationType.INVALID

    def _validate_bondlength(self) -> None:
        """Validate the bond length for the specified molecule and basis set.

        This method:
        1. Retrieves the list of valid bond lengths for the given molecule and basis set
           from PennyLane's dataset
        2. Checks if the provided bond length is in the list of valid values
        3. Raises ValueError if the bond length is not valid

        Raises:
            ValueError: If the provided bond length is not in the list of valid values
                       for the given molecule and basis set.
        """
        valid_bondlengths = qml.data.list_datasets()[DATA_MODULE_NAME][self.molname][self.basis_name]
        if str(self.bondlength) not in valid_bondlengths:
            raise ValueError(f"Invalid bondlength: {self.bondlength}. Valid bondlengths are {valid_bondlengths}")

    def _initialize_hamiltonian(self) -> None:
        """Initialize the molecular Hamiltonian.

        This method:
        1. Creates a Molecule object either from:
           - A predefined dataset (if molname and bondlength are provided)
           - Atomic symbols and coordinates (if symbols and coordinates are provided)
        2. Constructs the molecular Hamiltonian using PennyLane's quantum chemistry module
        3. Sets the number of qubits required for the simulation
        4. Raises ValueError if neither dataset nor atomic information is provided
        """
        init_type = self._determine_initialization_type()
        if init_type == InitializationType.DATASET:
            self._validate_bondlength()
            # force=False: if the dataset already exists, it will not be loaded again
            self._dataset = qml.data.load(
                DATA_MODULE_NAME, molname=self.molname, basis=self.basis_name, bondlength=self.bondlength, force=False
            )
            self.molecule = self._dataset[0].molecule
        elif init_type == InitializationType.DIRECT_MOLECULE:
            self.molecule = qml.qchem.Molecule(
                self.symbols,
                self.coordinates,
                charge=self.charge,
                mult=self.multi,
                basis_name=self.basis_name,
                unit=self.unit,
            )
        else:
            raise ValueError('Either "molname" and "bondlength" or "symbols" and "coordinates" must be provided')

        hamiltonian, n_qubits = qml.qchem.molecular_hamiltonian(
            molecule=self.molecule,
            method=self.method,
            active_electrons=self.active_electrons,
            active_orbitals=self.active_orbitals,
            mapping=self.mapping,
        )
        self.hamiltonian = hamiltonian
        self.n_qubits = n_qubits

    def _set_hf_energy(self) -> None:
        """Set the Hartree-Fock energy of the molecule.

        Returns:
            float: Hartree-Fock energy.
        """
        self.hf_energy = float(qml.qchem.hf_energy(self.molecule)())

    def _set_fci_energy(self) -> None:
        """Set the FCI energy of the molecule.
        FCI energy is only available for molecules in the dataset.
        If the molecule is directly specified, FCI energy is not available.

        [TODO]: FCI energy get from pyscf

        Returns:
            float: FCI energy.
        """
        if self.molname is not None:
            if not self._dataset:
                raise ValueError("Dataset is not loaded. Please load the dataset first.")
            self.fci_energy = float(self._dataset[0].fci_energy)
        else:
            self.logger.warning("FCI energy is not available for the given molecule. Setting to None.")
            self.fci_energy = None

    def get_hamiltonian(self) -> Sum:
        """Get the Hamiltonian operator.

        Returns:
            Sum: The molecular Hamiltonian operator.
        """
        return self.hamiltonian

    def get_n_qubits(self) -> int:
        """Get the number of qubits required for the simulation.

        Returns:
            int: Number of qubits.
        """
        return self.n_qubits

    def get_molecule(self) -> qml.qchem.Molecule:
        """Get the Molecule object.

        Returns:
            qml.qchem.Molecule: The molecule object.
        """
        return self.molecule

    def get_electrons(self) -> int:
        """Get the total number of electrons in the molecule.

        Returns:
            int: Total number of electrons.
        """
        return self.molecule.n_electrons

    def get_molecular_orbitals(self) -> int:
        """Get the number of molecular orbitals.

        Returns:
            int: Number of molecular orbitals.
        """
        return self.molecule.n_orbitals

    def get_spin_orbitals(self) -> int:
        """Get the number of spin orbitals.

        Returns:
            int: Number of spin orbitals (2 * number of molecular orbitals).
        """
        return 2 * self.molecule.n_orbitals

    def get_active_electrons(self) -> int:
        """Get the number of active electrons.

        If active_electrons is not specified, returns the total number of electrons.

        Returns:
            int: Number of active electrons.
        """
        if self.active_electrons is None:
            return self.get_electrons()
        else:
            return self.active_electrons

    def get_active_orbitals(self) -> int:
        """Get the number of active orbitals.

        If active_orbitals is not specified, returns the total number of molecular orbitals.

        Returns:
            int: Number of active orbitals.
        """
        if self.active_orbitals is None:
            return self.get_molecular_orbitals()
        else:
            return self.active_orbitals

    def get_active_spin_orbitals(self) -> int:
        """Get the number of active spin orbitals.

        If active_orbitals is not specified, returns the total number of spin orbitals.

        Returns:
            int: Number of active spin orbitals (2 * number of active orbitals).
        """
        if self.active_orbitals is None:
            return self.get_spin_orbitals()
        else:
            return 2 * self.active_orbitals

    def get_hf_energy(self) -> float:
        """Get the Hartree-Fock energy of the molecule.

        Returns:
            float: Hartree-Fock energy.
        """
        return self.hf_energy

    def get_fci_energy(self) -> Optional[float]:
        """Get the FCI energy of the molecule.

        Returns:
            float: FCI energy.
        """
        return self.fci_energy



from typing import Callable, cast

import numpy as np
import pennylane as qml
from pennylane.measurements.sample import SampleMP
from pennylane.measurements.state import StateMP
from rich.progress import track

from qxmt.devices import BaseDevice
from qxmt.exceptions import ModelSettingError
from qxmt.feature_maps import BaseFeatureMap
from qxmt.kernels import BaseKernel
from qxmt.kernels.sampling import sample_results_to_probs


class FidelityKernel(BaseKernel):
    """Fidelity kernel class.
    The fidelity kernel is a quantum kernel that computes the kernel value based on the fidelity
    between two quantum states.

    Args:
        BaseKernel (_type_): base class of kernel

    Examples:
        >>> import numpy as np
        >>> from qxmt.kernels.pennylane.fidelity_kernel import FidelityKernel
        >>> from qxmt.feature_maps.pennylane.defaults import ZZFeatureMap
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1024,
        >>> )
        >>> device = DeviceBuilder(config).build()
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> kernel = FidelityKernel(device, feature_map)
        >>> x1 = np.random.rand(2)
        >>> x2 = np.random.rand(2)
        >>> kernel.compute(x1, x2)
        0.14
    """

    def __init__(
        self,
        device: BaseDevice,
        feature_map: BaseFeatureMap | Callable[[np.ndarray], None],
    ) -> None:
        """Initialize the FidelityKernel class.

        Args:
            device (BaseDevice): device instance for quantum computation
            feature_map (BaseFeatureMap | Callable[[np.ndarray], None]): feature map instance or function
        """

        super().__init__(device, feature_map)
        self.qnode = None
        self.state_memory = {}

    def _initialize_qnode(self) -> None:
        if (self.qnode is None) and (self.is_sampling):
            self.qnode = qml.QNode(self._circuit, device=self.device.get_device(), cache=False, diff_method=None)
        elif (self.qnode is None) and (not self.is_sampling):
            self.qnode = qml.QNode(
                self._circuit_state_vector, device=self.device.get_device(), cache=False, diff_method=None
            )

    def _circuit(self, x1: np.ndarray, x2: np.ndarray) -> SampleMP | list[SampleMP]:
        if self.feature_map is None:
            raise ModelSettingError("Feature map must be provided for FidelityKernel.")

        self.feature_map(x1)
        qml.adjoint(self.feature_map)(x2)  # type: ignore

        if (self.is_sampling) and (self.device.is_amazon_device()):
            # Amazon Braket does not support directry sample by computational basis
            return [qml.sample(op=qml.PauliZ(wires=i)) for i in range(self.n_qubits)]
        else:
            return qml.sample(wires=range(self.n_qubits))

    def _circuit_state_vector(self, x: np.ndarray) -> StateMP:
        if self.feature_map is None:
            raise ModelSettingError("Feature map must be provided for FidelityKernel.")

        self.feature_map(x)

        return qml.state()

    def _compute_matrix_by_state_vector(
        self, x1_array: np.ndarray, x2_array: np.ndarray, bar_label: str = "", show_progress: bool = True
    ) -> np.ndarray:
        """Compute the kernel matrix based on the state vector.
        This method is only available in the non-sampling mode.
        Each kernel value computed by theoritically probability distribution by state vector.

        Args:
            x1_array (np.ndarray): numpy array representing the all data points (ex: Train data)
            x2_array (np.ndarray): numpy array representing the all data points (ex: Train data, Test data)
            bar_label (str): label for progress bar
            show_progress (bool): flag for showing progress bar

        Returns:
            np.ndarray: computed kernel matrix
        """
        self._initialize_qnode()
        if self.qnode is None:
            raise RuntimeError("QNode is not initialized.")

        unique_inputs = set([tuple(x) for x in x1_array] + [tuple(x) for x in x2_array])
        if show_progress:
            bar_label = f" ({bar_label})" if bar_label else ""
            iterator = track(unique_inputs, description=f"Computing Kernel Matrix{bar_label}")
        else:
            iterator = unique_inputs

        for x_tuple in iterator:
            if x_tuple not in self.state_memory:
                self.state_memory[x_tuple] = self.qnode(np.array(x_tuple))

        kernel_matrix = np.zeros((len(x1_array), len(x2_array)))
        for i, x1 in enumerate(x1_array):
            for j, x2 in enumerate(x2_array):
                kernel_matrix[i, j] = (
                    np.abs(np.dot(np.conj(self.state_memory[tuple(x1)]), self.state_memory[tuple(x2)])) ** 2
                )

        return kernel_matrix

    def _compute_by_sampling(self, x1: np.ndarray, x2: np.ndarray) -> tuple[float, np.ndarray]:
        """Compute the fidelity kernel value between two data points.
        This method is only available in the sampling mode.
        Each kernel value computed by sampling the quantum circuit.

        Args:
            x1 (np.ndarray): numpy array representing the first data point
            x2 (np.ndarray): numpy array representing the second data point

        Returns:
            tuple[float, np.ndarray]: fidelity kernel value and probability distribution
        """
        if not self.is_sampling:
            raise ValueError("_compute_by_sampling method is only available in sampling mode.")

        self._initialize_qnode()
        if self.qnode is None:
            raise RuntimeError("QNode is not initialized.")

        if (self.is_sampling) and (self.device.is_amazon_device()):
            result = self.qnode(x1, x2)
            # PauliZ basis convert to computational basis (-1->1, 1->0)
            binary_result = (np.array(result).T == -1).astype(int)
            # convert the sample results to probability distribution
            # shots must be over 0 when sampling mode
            probs = sample_results_to_probs(binary_result, self.n_qubits, cast(int, self.device.shots))
        else:
            result = self.qnode(x1, x2)
            # convert the sample results to probability distribution
            # shots must be over 0 when sampling mode
            probs = sample_results_to_probs(result, self.n_qubits, cast(int, self.device.shots))

        kernel_value = probs[0]  # get |0..0> state probability

        return kernel_value, probs



from qxmt.models.qkernels.base import BaseKernelModel, BaseMLModel

__all__ = ["BaseMLModel", "BaseKernelModel"]

from qxmt.models.qkernels.qrigge import QRiggeRegressor
from qxmt.models.qkernels.qsvc import QSVC
from qxmt.models.qkernels.qsvr import QSVR

__all__ += ["QSVC", "QSVR", "QRiggeRegressor"]

from qxmt.models.qkernels.builder import KernelModelBuilder

__all__ += ["KernelModelBuilder"]



from typing import Any, Optional

from qxmt.ansatze import BaseAnsatz
from qxmt.configs import AnsatzConfig
from qxmt.devices import BaseDevice
from qxmt.hamiltonians import BaseHamiltonian
from qxmt.utils import load_object_from_yaml


class AnsatzBuilder:
    def __init__(self, config: AnsatzConfig, device: BaseDevice, hamiltonian: Optional[BaseHamiltonian] = None):
        self.config = config
        self.device = device
        self.hamiltonian = hamiltonian

    def build(self) -> BaseAnsatz:
        dynamic_params: dict[str, Any] = {"device": self.device}

        if self.hamiltonian is not None:
            dynamic_params["hamiltonian"] = self.hamiltonian

        ansatz = load_object_from_yaml(
            config=self.config.model_dump(),
            dynamic_params=dynamic_params,
        )

        if not isinstance(ansatz, BaseAnsatz):
            raise TypeError("Ansatz must be a BaseAnsatz instance.")

        return ansatz



from qxmt.configs import HamiltonianConfig
from qxmt.hamiltonians import BaseHamiltonian
from qxmt.utils import load_object_from_yaml


class HamiltonianBuilder:
    def __init__(self, config: HamiltonianConfig):
        self.config = config

    def build(self) -> BaseHamiltonian:
        self.hamiltonian = load_object_from_yaml(config=self.config.model_dump())

        if not isinstance(self.hamiltonian, BaseHamiltonian):
            raise TypeError("Hamiltonian must be a BaseHamiltonian instance.")

        return self.hamiltonian



from logging import Logger
from typing import Any, Optional, cast

import numpy as np
import pennylane as qml
from pennylane.measurements import ExpectationMP
from pennylane.ops.op_math import Sum
from pennylane.workflow.qnode import SupportedDiffMethods

from qxmt.ansatze import BaseAnsatz
from qxmt.devices import BaseDevice
from qxmt.hamiltonians import BaseHamiltonian
from qxmt.logger import set_default_logger
from qxmt.models.vqe.base import BaseVQE, OptimizerPlatform

LOGGER = set_default_logger(__name__)


class BasicVQE(BaseVQE):
    """Basic implementation of the Variational Quantum Eigensolver (VQE).

    This class provides a basic implementation of VQE using PennyLane's optimization tools.
    It supports gradient-based optimization of quantum circuits to find the ground state
    energy of a given Hamiltonian. SciPy optimizers can now use PennyLane's automatic
    differentiation for efficient gradient computation.

    Args:
        device: Quantum device to use for the VQE.
        hamiltonian: Hamiltonian to find the ground state of.
        ansatz: Quantum circuit ansatz to use.
        diff_method: Method to use for differentiation. Defaults to "adjoint".
        max_steps: Maximum number of optimization steps. Defaults to 100.
        min_steps: Minimum number of optimization steps. Defaults to 1/10 of max_steps.
        tol: Tolerance for the optimization. Defaults to 1e-6.
        verbose: Whether to output progress during optimization. Defaults to True.
        optimizer_settings: Settings for the optimizer. For SciPy optimizers, you can
            set "gradient_method": "numerical" to use SciPy's numerical differentiation
            instead of PennyLane's automatic differentiation.
        logger: Logger object for output. Defaults to module-level logger.

    Example:
        Using SciPy optimizer with PennyLane automatic differentiation::

            optimizer_settings = {
                "name": "scipy.BFGS",
                "params": {
                    "gradient_method": "autodiff",  # Use PennyLane autodiff (default)
                    "options": {"gtol": 1e-6}
                }
            }
            vqe = BasicVQE(device, hamiltonian, ansatz, optimizer_settings=optimizer_settings)
            vqe.optimize()

        Using SciPy numerical differentiation::

            optimizer_settings = {
                "name": "scipy.BFGS",
                "params": {
                    "gradient_method": "numerical",  # Use SciPy numerical gradients
                }
            }

    Attributes:
        cost_history: List of cost values during optimization.
        params_history: List of parameter values during optimization.
    """

    def __init__(
        self,
        device: BaseDevice,
        hamiltonian: BaseHamiltonian,
        ansatz: BaseAnsatz,
        diff_method: Optional[SupportedDiffMethods] = "adjoint",
        max_steps: int = 100,
        min_steps: Optional[int] = None,
        tol: float = 1e-6,
        verbose: bool = True,
        optimizer_settings: Optional[dict[str, Any]] = None,
        init_params_config: Optional[dict[str, Any]] = None,
        logger: Logger = LOGGER,
    ) -> None:
        super().__init__(
            device=device,
            hamiltonian=hamiltonian,
            ansatz=ansatz,
            diff_method=diff_method,
            max_steps=max_steps,
            min_steps=min_steps,
            tol=tol,
            verbose=verbose,
            optimizer_settings=optimizer_settings,
            init_params_config=init_params_config,
            logger=logger,
        )

    def _initialize_qnode(self) -> None:
        """Initialize the QNode for VQE.

        This method creates a quantum node that:
        1. Executes the ansatz circuit with given parameters
        2. Measures the expectation value of the Hamiltonian

        Raises:
            ValueError: If the Hamiltonian is not a Sum instance.
        """

        def circuit_with_measurement(params: qml.numpy.ndarray | np.ndarray) -> ExpectationMP:
            self.ansatz.circuit(params)
            if not isinstance(self.hamiltonian.hamiltonian, Sum):
                raise ValueError("Hamiltonian must be a Sum instance.")
            else:
                return qml.expval(self.hamiltonian.hamiltonian)

        self.qnode = qml.QNode(
            func=circuit_with_measurement,
            device=self.device.get_device(),
            diff_method=cast(SupportedDiffMethods, self.diff_method),
        )

    def _optimize_scipy(self, init_params: np.ndarray) -> None:
        """Optimize the ansatz parameters using scipy.

        Args:
            init_params (np.ndarray): Initial parameters for the ansatz.
        """
        step_num = {"step": 0}

        def cost_function(params):
            cost = self.qnode(params)
            return float(cost)

        def callback(params):
            cost = self.qnode(params)
            self.cost_history.append(float(cost))
            self.params_history.append(params.copy())
            step_num["step"] += 1
            if self.verbose:
                self.logger.info(f"Step {step_num['step']}: Cost = {cost}")

        self.optimizer(
            init_params,
            cost_function,
            callback=callback,
        )

    def _optimize_pennylane(self, init_params: qml.numpy.ndarray) -> None:
        """Optimize the ansatz parameters using Pennylane.

        Args:
            init_params (qml.numpy.ndarray): Initial parameters for the ansatz.
        """
        params = init_params
        for i in range(self.max_steps):
            params, cost = self.optimizer.step_and_cost(self.qnode, params)
            self.cost_history.append(cost)
            self.params_history.append(params)
            if self.verbose:
                self.logger.info(f"Step {i+1}: Cost = {cost}")

            if i > self.min_steps and abs(self.cost_history[-1] - self.cost_history[-2]) < self.tol:
                self.logger.info(f"Optimization finished at step {i+1}.")
                break

    def optimize(self, init_params: Optional[qml.numpy.ndarray | np.ndarray] = None) -> None:
        """Optimize the ansatz parameters to find the ground state.

        This method performs gradient-based optimization of the ansatz parameters
        to minimize the expectation value of the Hamiltonian.

        Args:
            init_params (Optional[qml.numpy.ndarray | np.ndarray]): Initial parameters for the ansatz.
                If None, the ansatz parameters are initialized to zero.

        Note:
            The optimization history (cost and parameters) is stored in the class attributes
            cost_history and params_history.
        """
        self._set_optimizer()
        self.logger.info(f"Optimizing ansatz with {self.ansatz.n_params} parameters through {self.max_steps} steps")

        if self.optimizer_platform == OptimizerPlatform.SCIPY:
            if init_params is None:
                init_params = self._parse_init_params(self.init_params_config, self.ansatz.n_params)
            self._optimize_scipy(init_params)
        else:
            if init_params is None:
                init_params = self._parse_init_params(self.init_params_config, self.ansatz.n_params)
            self._optimize_pennylane(init_params)

        self.logger.info(f"Optimization finished. Final cost: {self.cost_history[-1]:.8f}")



from qxmt.configs import ExperimentConfig
from qxmt.constants import DEFAULT_N_JOBS
from qxmt.models.qkernels import BaseMLModel, KernelModelBuilder
from qxmt.models.vqe import BaseVQE, VQEModelBuilder


class ModelBuilder:
    """Builder class for quantum machine learning models.
    This class is responsible for building quantum machine learning models.
    """

    def __init__(self, config: ExperimentConfig, n_jobs: int = DEFAULT_N_JOBS, show_progress: bool = True) -> None:
        self.config = config
        self.n_jobs = n_jobs
        self.show_progress = show_progress

    def build(self) -> BaseMLModel | BaseVQE:
        model_type = self.config.global_settings.model_type
        match model_type:
            case "qkernel":
                return KernelModelBuilder(self.config, self.n_jobs, self.show_progress).build()
            case "vqe":
                return VQEModelBuilder(self.config, self.n_jobs).build()
            case _:
                raise ValueError(f"Invalid model type: {model_type}")



from typing import Any, Literal, Type

from qxmt.evaluation.metrics.base import BaseMetric
from qxmt.hamiltonians.pennylane.molecular import MolecularHamiltonian


class FinalCost(BaseMetric):
    """Metric for calculating the final cost from VQE optimization history.

    This metric extracts the last cost value from the cost history, representing
    the final cost obtained from the VQE algorithm.
    """

    def __init__(self, name: str = "final_cost") -> None:
        super().__init__(name)

    @staticmethod
    def evaluate(cost_history: list[float], **kwargs: Any) -> float:
        """Calculate the final cost.

        Args:
            cost_history (list[float]): List of cost values from VQE optimization

        Returns:
            float: The final cost value
        """
        return cost_history[-1]


class HFEnergy(BaseMetric):
    """Metric for calculating the Hartree-Fock (HF) energy.

    This metric computes the Hartree-Fock energy using the molecular Hamiltonian.
    """

    def __init__(self, name: str = "hf_energy") -> None:
        super().__init__(name)

    @staticmethod
    def evaluate(hamiltonian: MolecularHamiltonian, **kwargs: Any) -> float:
        """Get the precomputed Hartree-Fock energy.

        Args:
            hamiltonian (MolecularHamiltonian): The molecular Hamiltonian object

        Returns:
            float: The Hartree-Fock energy value
        """
        return hamiltonian.get_hf_energy()


class FCIEnergy(BaseMetric):
    """Metric for calculating the Full Configuration Interaction (FCI) energy.

    This metric computes the FCI energy using the molecular Hamiltonian.
    The FCI energy represents the exact solution within the given basis set.
    """

    def __init__(self, name: str = "fci_energy") -> None:
        super().__init__(name)

    @staticmethod
    def evaluate(hamiltonian: MolecularHamiltonian, **kwargs: Any) -> float | None:
        """Get the precomputed Full Configuration Interaction energy.

        Args:
            hamiltonian (MolecularHamiltonian): The molecular Hamiltonian object

        Returns:
            float | None: The FCI energy value, or None if not available
        """
        return hamiltonian.get_fci_energy()


# set default vqe metrics name list for evaluation
DEFAULT_VQE_METRICS_NAME = Literal[
    "final_cost",
    "hf_energy",
    "fci_energy",
]


NAME2VQE_METRIC: dict[str, Type[BaseMetric]] = {
    "final_cost": FinalCost,
    "hf_energy": HFEnergy,
    "fci_energy": FCIEnergy,
}



from abc import ABC, abstractmethod
from enum import Enum
from logging import Logger
from typing import Any, Optional

import numpy as np
import pennylane as qml
from pennylane.workflow.qnode import QNode, SupportedDiffMethods

from qxmt.ansatze import BaseAnsatz
from qxmt.devices import BaseDevice
from qxmt.hamiltonians import BaseHamiltonian
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)
DEFAULT_OPTIMIZER_STEPSIZE = 0.5
SCIPY_OPTIMIZER_PREFIX = "scipy."
INIT_PARAMS_TYPE_ZEROS = "zeros"
INIT_PARAMS_TYPE_RANDOM = "random"
INIT_PARAMS_TYPE_CUSTOM = "custom"


class OptimizerPlatform(Enum):
    SCIPY = "scipy"
    PENNYLANE = "pennylane"


class BaseVQE(ABC):
    """Base class for Variational Quantum Eigensolver (VQE).

    This abstract base class provides the foundation for implementing VQE algorithms.
    It handles the setup of quantum circuits, parameter optimization, and tracking of
    optimization history.

    Args:
        device: Quantum device to use for the VQE calculation.
        hamiltonian: Hamiltonian to find the ground state of.
        ansatz: Quantum circuit ansatz to use.
        max_steps: Maximum number of optimization steps. Defaults to 100.
        min_steps: Minimum number of optimization steps. Defaults to 1/10 of max_steps.
        tol: Tolerance for the optimization. Defaults to 1e-6.
        verbose: Whether to output progress during optimization. Defaults to True.
        diff_method: Method to use for differentiation. Defaults to "adjoint".
        optimizer_settings: Settings for the optimizer.
        logger: Logger object for output. Defaults to module-level logger.

    Attributes:
        qnode: Quantum node for executing the circuit.
        params_history: List of parameter values during optimization.
        cost_history: List of cost values during optimization.
    """

    def __init__(
        self,
        device: BaseDevice,
        hamiltonian: BaseHamiltonian,
        ansatz: BaseAnsatz,
        diff_method: Optional[SupportedDiffMethods] = "adjoint",
        max_steps: int = 100,
        min_steps: Optional[int] = None,
        tol: float = 1e-6,
        verbose: bool = True,
        optimizer_settings: Optional[dict[str, Any]] = None,
        init_params_config: Optional[dict[str, Any]] = None,
        logger: Logger = LOGGER,
    ) -> None:
        self.device = device
        self.hamiltonian = hamiltonian
        self.ansatz = ansatz
        self.diff_method: Optional[SupportedDiffMethods] = diff_method
        self.max_steps: int = max_steps
        self.min_steps: int = min_steps if min_steps is not None else int(max_steps * 0.1)
        self.tol: float = float(tol)
        self.verbose: bool = verbose
        self.optimizer_settings: Optional[dict[str, Any]] = optimizer_settings
        self.init_params_config: Optional[dict[str, Any]] = init_params_config
        self.logger: Logger = logger
        self.qnode: QNode
        self.optimizer: Any
        self.params_history: list[np.ndarray] = []
        self.cost_history: list[float] = []
        self._set_optimizer_platform()
        self._initialize_qnode()

    @abstractmethod
    def _initialize_qnode(self) -> None:
        """Initialize the quantum node for VQE.

        This method must be implemented by subclasses to set up the quantum node
        that will be used for circuit execution and optimization.

        Note:
            The implementation should create a QNode that combines the ansatz circuit
            with the Hamiltonian measurement.
        """
        pass

    @abstractmethod
    def optimize(self, init_params: Any) -> None:
        """Optimize the ansatz parameters to find the ground state.

        This method must be implemented by subclasses to perform the parameter
        optimization process.

        Args:
            init_params: Initial parameters for the ansatz. This array must be calculate gradientable
                with the optimizer.

        Note:
            The implementation should update params_history and cost_history
            during the optimization process.
        """
        pass

    def _set_optimizer_platform(self) -> None:
        """Set the optimizer platform based on the optimizer settings."""
        optimizer_name = self.optimizer_settings.get("name", "") if self.optimizer_settings else ""
        if optimizer_name.startswith(SCIPY_OPTIMIZER_PREFIX):
            self.optimizer_platform = OptimizerPlatform.SCIPY
        else:
            self.optimizer_platform = OptimizerPlatform.PENNYLANE

    def _parse_init_params(self, init_params_config: Optional[dict[str, Any]], n_params: int) -> np.ndarray:
        """Parse the initial parameters based on the init_params_config.

        Args:
            init_params_config: Configuration for the initial parameters. If None, the default is zeros.
            n_params: Number of parameters in the ansatz.

        Returns:
            Initial parameters for the ansatz.
        """
        if init_params_config is None or init_params_config.get("type") == INIT_PARAMS_TYPE_ZEROS:
            return (
                np.zeros(n_params)
                if self.optimizer_platform == OptimizerPlatform.SCIPY
                else qml.numpy.zeros(n_params, requires_grad=True)
            )
        elif init_params_config.get("type") == INIT_PARAMS_TYPE_RANDOM:
            seed = init_params_config.get("random_seed", None)
            rng = np.random.default_rng(seed)
            return (
                rng.random(n_params)
                if self.optimizer_platform == OptimizerPlatform.SCIPY
                else qml.numpy.array(rng.random(n_params), requires_grad=True)
            )
        elif init_params_config.get("type") == INIT_PARAMS_TYPE_CUSTOM:
            values = init_params_config.get("values", None)
            if values is None or len(values) != n_params:
                raise ValueError("Custom init_params must provide a list of length n_params")
            return (
                np.array(values)
                if self.optimizer_platform == OptimizerPlatform.SCIPY
                else qml.numpy.array(values, requires_grad=True)
            )
        else:
            raise ValueError(f"Unknown init_params type: {init_params_config.get('type')}")

    def is_params_updated(self, threshold: float = 1e-8) -> bool:
        """Check if parameters were actually updated during optimization.

        Args:
            threshold: Minimum change threshold to consider parameters as updated.

        Returns:
            bool: True if parameters changed more than threshold, False otherwise.
        """
        if len(self.params_history) < 2:
            return False

        # Compare first and last parameters
        initial_params = self.params_history[0]
        final_params = self.params_history[-1]

        # Calculate L2 norm of parameter change
        param_change = float(np.linalg.norm(final_params - initial_params))

        return param_change > threshold

    def is_optimized(self) -> bool:
        """Check if the optimization process has been completed.

        Returns:
            bool: True if optimization has been performed (cost_history is not empty),
                 False otherwise.

        Note:
            This is a simple check that only verifies if any optimization steps
            have been taken. It does not check if the optimization has converged
            to a satisfactory solution.
        """
        return len(self.cost_history) > 0

    def _set_optimizer(self) -> None:
        """Set the optimizer based on the optimizer settings.

        This method sets the optimizer based on the optimizer settings.
        If no optimizer settings are provided, it uses the default optimizer.
        Otherwise, it uses the optimizer specified in the settings.
        """
        if self.optimizer_settings is None:
            self._set_default_optimizer()
            return

        optimizer_name = self.optimizer_settings.get("name", "")
        optimizer_params = self.optimizer_settings.get("params", {}) or {}

        if optimizer_name.startswith("scipy."):
            optimizer_params.setdefault("options", {})
            optimizer_params["options"]["maxiter"] = self.max_steps
            optimizer_params["tol"] = self.tol
            self._set_scipy_optimizer(optimizer_name, optimizer_params)
        else:
            self._set_pennylane_optimizer(optimizer_name, optimizer_params)

    def _set_default_optimizer(self) -> None:
        """Set the default gradient descent optimizer."""
        self.optimizer = qml.GradientDescentOptimizer(stepsize=DEFAULT_OPTIMIZER_STEPSIZE)
        self.logger.info("No optimizer settings provided. Using gradient descent optimizer.")

    def _set_scipy_optimizer(self, optimizer_name: str, optimizer_params: dict) -> None:
        """Set up a SciPy optimizer with PennyLane gradients.

        This method sets up SciPy optimizers to use PennyLane's automatic differentiation
        for gradient computation instead of SciPy's numerical differentiation.
        """
        from scipy.optimize import minimize

        method = optimizer_name.replace("scipy.", "")

        # Check gradient computation method (default: "autodiff")
        gradient_method = optimizer_params.pop("gradient_method", "autodiff")

        def scipy_optimizer(params, cost_fn, **kwargs):
            if gradient_method == "autodiff":
                # Create gradient function using PennyLane's automatic differentiation
                grad_fn = qml.grad(self.qnode)

                def compute_gradient(params_array):
                    """Wrapper function for gradient computation."""
                    # Convert params to qml.numpy array with requires_grad for gradient computation
                    qml_params = qml.numpy.array(params_array, requires_grad=True)
                    grad = grad_fn(qml_params)
                    return np.array(grad, dtype=float)

                gradient_func = compute_gradient
            else:
                # Use SciPy's numerical differentiation (gradient_method == "numerical")
                gradient_func = None

            return minimize(
                x0=params,
                fun=cost_fn,
                jac=gradient_func,  # Use PennyLane gradients or None for numerical diff
                method=method,
                **optimizer_params,
                **kwargs,
            ).x

        self.optimizer = scipy_optimizer

    def _set_pennylane_optimizer(self, optimizer_name: str, optimizer_params: dict) -> None:
        """Set up a Pennylane optimizer."""
        match optimizer_name:
            case "AdagradOptimizer" | "Adagrad":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.AdagradOptimizer.html
                self.optimizer = qml.AdagradOptimizer(**optimizer_params)
            case "AdamOptimizer" | "Adam":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.AdamOptimizer.html
                self.optimizer = qml.AdamOptimizer(**optimizer_params)
            case "AdaptiveOptimizer" | "Adaptive":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.AdaptiveOptimizer.html
                self.optimizer = qml.AdaptiveOptimizer(**optimizer_params)
            case "GradientDescentOptimizer" | "GradientDescent":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.GradientDescentOptimizer.html
                self.optimizer = qml.GradientDescentOptimizer(**optimizer_params)
            case "MomentumOptimizer" | "Momentum":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.MomentumOptimizer.html
                self.optimizer = qml.MomentumOptimizer(**optimizer_params)
            case "MomentumQNGOptimizer" | "MomentumQNG":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.MomentumQNGOptimizer.html
                self.optimizer = qml.MomentumQNGOptimizer(**optimizer_params)
            case "NesterovMomentumOptimizer" | "NesterovMomentum":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.NesterovMomentumOptimizer.html
                self.optimizer = qml.NesterovMomentumOptimizer(**optimizer_params)
            case "QNGOptimizer" | "QNG":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.QNGOptimizer.html
                self.optimizer = qml.QNGOptimizer(**optimizer_params)
            case "QNSPSAOptimizer" | "QNSPSA":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.QNSPSAOptimizer.html
                self.optimizer = qml.QNSPSOptimizer(**optimizer_params)
            case "RMSPropOptimizer" | "RMSProp":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.RMSPropOptimizer.html
                self.optimizer = qml.RMSPropOptimizer(**optimizer_params)
            case "RiemannianGradientOptimizer" | "RiemannianGradient":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.RiemannianGradientOptimizer.html
                self.optimizer = qml.RiemannianGradientOptimizer(**optimizer_params)
            case "RotoselectOptimizer" | "Rotoselect":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.RotoselectOptimizer.html
                self.optimizer = qml.RotoselectOptimizer(**optimizer_params)
            case "RotosolveOptimizer" | "Rotosolve":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.RotosolveOptimizer.html
                self.optimizer = qml.RotosolveOptimizer(**optimizer_params)
            case "SPSAOptimizer" | "SPSA":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.SPSAOptimizer.html
                self.optimizer = qml.SPSAOptimizer(**optimizer_params)
            case "ShotAdaptiveOptimizer" | "ShotAdaptive":
                # https://docs.pennylane.ai/en/stable/code/api/pennylane.ShotAdaptiveOptimizer.html
                self.optimizer = qml.ShotAdaptiveOptimizer(**optimizer_params)
            case _:
                raise NotImplementedError(f'Optimizer "{optimizer_name}" is not implemented yet.')



from qxmt.models.vqe.base import BaseVQE

__all__ = ["BaseVQE"]

from qxmt.models.vqe.basic import BasicVQE
from qxmt.models.vqe.builder import VQEModelBuilder

__all__ += ["BasicVQE", "VQEModelBuilder"]



from datetime import datetime
from typing import Any, Optional

import numpy as np
import pennylane as qml

from qxmt.devices.base import BaseDevice
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


class PennyLaneDevice(BaseDevice):
    """PennyLane device implementation for quantum computation.
    This class provides a concrete implementation for PennyLane devices.
    """

    def __init__(
        self,
        platform: str,
        device_name: str,
        backend_name: Optional[str],
        n_qubits: int,
        shots: Optional[int],
        random_seed: Optional[int] = None,
        logger: Any = LOGGER,
    ) -> None:
        """Initialize the PennyLane device.

        Args:
            platform (str): platform name (ex: pennylane, qulacs, etc.)
            device_name (str): device name provided by the platform (ex: default.qubit, default.tensor, etc.)
            backend_name (Optional[str]): backend name for the real device
            n_qubits (int): number of qubits
            shots (Optional[int]): number of shots for the quantum circuit
            random_seed (Optional[int]): random seed for the quantum device
            logger (Any): logger instance
        """
        super().__init__(platform, device_name, backend_name, n_qubits, shots, random_seed, logger)
        self.real_device = None

    def get_device(self) -> Any:
        """Get the quantum device instance.

        Returns:
            Any: quantum device instance
        """
        return qml.device(
            name=self.device_name,
            wires=self.n_qubits,
            shots=self.shots,
            seed=np.random.default_rng(self.random_seed) if self.random_seed is not None else None,
        )

    def is_simulator(self) -> bool:
        """Check if the device is a simulator or real machine.

        Returns:
            bool: True if the device is a simulator, False otherwise
        """
        return True

    def is_remote(self) -> bool:
        """Check if the device is a remote device.

        Returns:
            bool: True if the device is a remote device, False otherwise
        """
        return False

    def get_provider(self) -> str:
        """Get real machine provider name.

        Returns:
            str: provider name (empty for non-remote devices)
        """
        return ""

    def get_backend_name(self) -> str:
        """Get real machine backend name.

        Returns:
            str: backend name (empty for non-remote devices)
        """
        return ""

    def get_job_ids(
        self, created_after: Optional[datetime] = None, created_before: Optional[datetime] = None
    ) -> list[str]:
        """Get the job IDs.
        Local machine does not have job IDs.

        Args:
            created_after (Optional[datetime]): created datetime of the jobs. If None, start time filter is not applied.
            created_before (Optional[datetime]): finished datetime of the jobs. If None, end time filter is not applied.

        Returns:
            list[str]: job IDs (empty for non-remote devices)
        """
        return []



from abc import ABC, abstractmethod
from typing import Any


class BaseHamiltonian(ABC):
    """Base class for quantum Hamiltonians.

    This abstract base class defines the interface for Hamiltonian representations
    in quantum computing. It provides a common structure for different types of
    Hamiltonians across various quantum computing platforms.

    Args:
        platform: Name of the quantum computing platform (e.g., "pennylane").

    Attributes:
        hamiltonian: The Hamiltonian operator. Type depends on the platform.
        n_qubits: Number of qubits required to represent the Hamiltonian.
    """

    def __init__(self, platform: str) -> None:
        self.platform: str = platform
        self.hamiltonian: Any | None = None
        self.n_qubits: int | None = None

    @abstractmethod
    def get_hamiltonian(self) -> Any:
        """Get the Hamiltonian operator.

        This method must be implemented by subclasses to return the Hamiltonian
        operator in the format required by the specific platform.

        Returns:
            Any: The Hamiltonian operator. The exact type depends on the platform.

        Note:
            The returned operator should be compatible with the quantum computing
            platform specified in the constructor.
        """
        pass

    @abstractmethod
    def get_n_qubits(self) -> int:
        """Get the number of qubits required to represent the Hamiltonian.

        This method must be implemented by subclasses to return the number of
        qubits needed to represent the Hamiltonian on a quantum computer.

        Returns:
            int: Number of qubits required.

        Note:
            This value is used to validate that the quantum device has sufficient
            qubits to handle the Hamiltonian.
        """
        pass



from typing import cast

import numpy as np
import pennylane as qml

from qxmt.ansatze import BaseVQEAnsatz
from qxmt.devices import BaseDevice
from qxmt.hamiltonians.pennylane.molecular import MolecularHamiltonian


class UCCSDAnsatz(BaseVQEAnsatz):
    """Unitary Coupled-Cluster Singles and Doubles (UCCSD) ansatz.

    This class implements the UCCSD ansatz for quantum chemistry calculations.
    UCCSD is a popular ansatz for quantum chemistry that includes single and double excitations
    from the Hartree-Fock reference state.

    Args:
        device (BaseDevice): Quantum device to use for the ansatz.
        hamiltonian (MolecularHamiltonian): Molecular Hamiltonian to use for the ansatz.

    Attributes:
        wires: List of qubit indices used in the circuit.
        hf_state: Hartree-Fock reference state.
        singles: List of single excitation indices.
        doubles: List of double excitation indices.
        s_wires: List of wire indices for single excitations.
        d_wires: List of wire indices for double excitations.

    Example:
        >>> from qxmt.devices.pennylane import DefaultQubit
        >>> from qxmt.hamiltonians.pennylane.molecular import MolecularHamiltonian
        >>> device = DefaultQubit(n_qubits=4)
        >>> hamiltonian = MolecularHamiltonian(...)
        >>> ansatz = UCCSDAnsatz(device, hamiltonian)
        >>> params = np.random.rand(10)  # Number of parameters depends on the system
        >>> ansatz.circuit(params)

    Note:
        The number of parameters required depends on the number of single and double excitations,
        which is determined by the number of electrons and orbitals in the system.

    URL: https://docs.pennylane.ai/en/stable/code/api/pennylane.UCCSD.html
    """

    def __init__(self, device: BaseDevice, hamiltonian: MolecularHamiltonian) -> None:
        super().__init__(device, hamiltonian)
        self.hamiltonian = cast(MolecularHamiltonian, self.hamiltonian)
        self.wires = range(self.hamiltonian.get_n_qubits())
        self.prepare_hf_state()
        self.prepare_excitation_wires()
        self.n_params = len(self.singles) + len(self.doubles)

    def prepare_hf_state(self) -> None:
        """Prepare the Hartree-Fock reference state.

        This method creates the Hartree-Fock reference state using PennyLane's qchem module.
        The Hartree-Fock state is a product state where the first n electrons occupy the lowest
        energy orbitals.

        The state is stored in self.hf_state as a numpy array.

        Note:
            The number of electrons and orbitals are obtained from the Hamiltonian.
        """
        self.hf_state = qml.qchem.hf_state(
            electrons=self.hamiltonian.get_active_electrons(), orbitals=self.hamiltonian.get_active_spin_orbitals()
        )

    def prepare_excitation_wires(self) -> None:
        """Prepare the excitation wires for single and double excitations.

        This method:
        1. Generates all possible single and double excitations from the Hartree-Fock state
        2. Converts these excitations to wire indices that can be used in the quantum circuit

        The results are stored in the following attributes:
        - self.singles: List of single excitation indices
        - self.doubles: List of double excitation indices
        - self.s_wires: List of wire indices for single excitations
        - self.d_wires: List of wire indices for double excitations

        Note:
            The number of excitations depends on the number of electrons and orbitals in the system.
        """
        self.singles, self.doubles = qml.qchem.excitations(
            electrons=self.hamiltonian.get_active_electrons(), orbitals=self.hamiltonian.get_active_spin_orbitals()
        )
        self.s_wires, self.d_wires = qml.qchem.excitations_to_wires(self.singles, self.doubles)

    def circuit(self, params: np.ndarray) -> None:
        """Construct the UCCSD quantum circuit.

        Args:
            params (np.ndarray): Parameters for the UCCSD circuit. The length of this array
                               should match the number of single and double excitations.

        Note:
            The UCCSD operation includes both single and double excitations from the
            Hartree-Fock reference state.
        """
        qml.UCCSD(params, self.wires, self.s_wires, self.d_wires, self.hf_state)
